// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------
//
//                     XMM Science Analysis System (SAS)
//                    (c) 1997-2010 European Space Agency
//
//---------------------------------------------------------------------------
//
//	ArrayEvaluator.cc:	--- implementation of ArrayEvaluator class
//
//	CHANGE HISTORY:
//		UL, 1998-05-27: 1.0.0: - first version
//		UL, 1999-03-03: 1.1.0: - changes enforced by Lowmem Dal
//		UL, 1999-11-03: 1.2.0: - added support for new unsigned types in
//								 dal>=1.80
//		UL, 2002-07-19: 1.3.0: - minor C++ standard compliance fix w.r.t
//								 preprocessor usage
//
//	AUTHOR:
//		U. Lammers (UL), Wed May 27 18:53:51 MET DST 1998
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/selectlib/src/service/ArrayEvaluator.cc,v 1.4 2009/06/26 15:51:59 aibarra Exp $
//

#include <cstdio>	//	tmpnam()
/* AI gcc4 portin
 * memcpy is now in cstring
 */
//#include <cstdlib>	//	memcpy()
#include <cstring>
#include "../selcore/basicdefs.h"
#include "ArrayEvaluator.h"
#include "ServiceUtils.h"
#include "SelectError.h"

#if !defined(NON_SAS)

using namespace SelectError;
using namespace std;

static void copyData(Data *to, const Data *from, unsigned n,Dal::DataType type)
{
	#define COPY(t1, t2) \
		case Dal:: t1 : std::memcpy(to->t2##Data(), from->t2##Data(), \
				n * sizeof(t2)); break

	switch (type) {
		COPY(Int8, int8);
		COPY(Int16, int16);
		COPY(Uint16, uint16);
		COPY(Int32, int32);
		COPY(Uint32, uint32);
		COPY(Real32, real32);
		COPY(Real64, real64);
		COPY(Bool, bool);
		default: selerrstr<<"Unsupported data type "<<UNT(type)
						  <<" in copy data method encountered"
						  <<selerror("UnsupportedDataType");
				 SELSTDTHROW;
				 break;
	}

	#undef COPY
} // copyData()


static void addArrayColumn(Table &table, const std::string &arrayColName,
						   const Array &array)
{
	if (array.elements() != table.rows()) {
		selerrstr<<"Array `"<<array.qualifiedName()<<"' does not have right "
				   "dimensionality and/or extents\n"
				 <<selerror("WrongArrayDimensions");
		SELSTDTHROW;
	}

	Column *newCol = table.addColumn(arrayColName, array.dataType(), "");

	//	copy the data
	ColumnData *cd = newCol->data();
	ArrayData *ad = array.data();
	copyData(cd, ad, table.rows(), array.dataType());
	delete cd;
	delete ad;
} // addArrayColumn()


//	construct filter with boolean expression
ArrayEvaluator::ArrayEvaluator(const std::string &expression)
		: _expression(expression), _columnGenerator(expression, "RESULT")
{
} // ArrayEvaluator::ArrayEvaluator()


void ArrayEvaluator::doIt(const std::vector<Array *> arrayVec, Array &outArray)
{
	if (arrayVec.size() == 0)
		return;	//	nothing to do

	Evaluator eval(_expression);
	NumericSymbols idents = eval.numericIdentifiers();
	unsigned nIds = idents.name.size();
	if (nIds != arrayVec.size()) {
		selerrstr<<"Number of identifiers in expression ("<<nIds
				 <<") does not coincide with number of arrays given ("
				 <<arrayVec.size()<<")"<<selerror("NumberMismatch");
		SELSTDTHROW;
	}

	//	also using row numbers is not meaningful
	if (eval.rowNumber()) {
		selerrstr<<"Usage of row number identifier in array expression is "
				   "not meaningful"<<selerror("MeaninglessOperation");
		SELSTDTHROW;
	}

	//	let's do it
	DataSet *set = dataSetServer->open(std::tmpnam(0), DataSet::Temp);
	Table *table = set->addTable("ARRAYTAB", outArray.elements());
	for (unsigned n=0; n<nIds; ++n)
		addArrayColumn(*table, idents.name[n], *arrayVec[n]);

	//	generate result column
	_columnGenerator.setDataType(outArray.dataType());
	_columnGenerator.doIt(*table, *table);

	//	copy output to array data section
	ColumnData *cd = table->column(_columnGenerator.columnName())->data();
	ArrayData *ad = outArray.data();
	copyData(ad, cd, table->rows(), outArray.dataType());

	//	tidy up
	delete cd;
	delete ad;
	dataSetServer->close(set);
}

#endif
